unit sky_toast2;
{fits reading and mouse action on window realsky}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, math;

type
  TRealsky = class(TForm)
    Image1: TImage;
    StatusBar1: TStatusBar;
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);

  private
    { Private declarations }
  public
    { Public declarations }

  end;

var
  Realsky: TRealsky;


function prepare_ra(rax:double; sep:char):ansistring; {radialen to text, format 24: 00 00.0 }
function prepare_dec(decx:double; sep:char):ansistring; {radialen to text, format 90d 00 00}
procedure loadfits(filen:ansistring);{load fits file, by Han Kleijn}
procedure plotfits(filen:ansistring);{plot FITS scaled in 360x180 degrees Equirectangular projection}
procedure de_compensate_strech(x,y: double;var x2,y2:double); {correct mouse position for possible stretched image}

const
   nrbits    : integer=8;
   equinox   : double=2000.0;
   naxis     : integer=2; 
   wide2     : integer=100;
   height2   : integer=100;{do not use reserved word height}
const
   crpix1: double=0;{reference pixel}
   crpix2: double=0;

var
  angle     : double; {image rotation at center, radians}
  cdelt1    : double;{deg/pixel for x}
  cdelt2    : double;
  ra0,dec0  : double;
  crota2      : double; {image rotation at reference center in degrees}



implementation

uses sky_toast1;

{$R *.DFM}


var
  TheFile3  : tfilestream;
  Reader    : TReader;
//  header    : array[0..$4000] of char;
  header    : array[0..$4000] of ansichar;

const   bufwide=$4000; {16384}
type  byteX3 = array[0..2] of byte;
var
  fitsbuffer : array[0..bufwide] of byte;{buffer for 8 bit FITS file}
  fitsbufferRGB: array[0..trunc(bufwide/3)] of byteX3 absolute fitsbuffer;{buffer for 8 bit RGB FITS file}
  fitsbuffer2: array[0..trunc(bufwide/2)] of word absolute fitsbuffer;{buffer for 16 bit FITS file}
  fitsbuffer4: array[0..trunc(bufwide/4)] of longword absolute fitsbuffer;{buffer for floating bit ( -32) FITS file}
  fitsbuffer8: array[0..trunc(bufwide/8)] of int64 absolute fitsbuffer;{buffer for floating bit ( -64) FITS file}

  ram,decm  : double; {mouse position}
  cd1_1,cd1_2,cd2_1,cd2_2 :double;



(*-----------------------------------------------------------------------*)
(* ATN_2: arctangent of y/x for two arguments                             *)
(*       (correct quadrant; -pi rad <= ATN2 <= +pi rad)                  *)
(*-----------------------------------------------------------------------*)
FUNCTION ATN_2(Y,X:double):double;
  VAR   AX,AY,PHI: double;
  BEGIN
    IF (X=0.0) AND (Y=0.0)
      THEN ATN_2:=0.0
      ELSE
        BEGIN
          AX:=ABS(X); AY:=ABS(Y);
          IF (AX>AY)
            THEN PHI:=ARCTAN(AY/AX)
            ELSE PHI:=(pi/2)-ARCTAN(AX/AY);
          IF (X<0.0) THEN PHI:=pi-PHI;
          IF (Y<0.0) THEN PHI:=-PHI;
          ATN_2:=PHI;
        END;
  END;


Function LeadingZero(w : integer) : ansistring;
 var
   s : ansistring;
 begin
   Str(w:0,s);
   if Length(s) = 1 then
     s := '0' + s;
   LeadingZero := s;
 end;


function prepare_ra(rax:double; sep:char):ansistring; {radialen to text, format 24: 00 00.0 }
 var
   B : string[2];
   h,m,s,ds  :integer;
 begin {make from rax [0..pi*2] a text in array bericht. Length is 8 long}
  rax:=rax+pi*0.1/(24*60*60); {add 1/10 of half second to get correct rounding and not 7:60 results as with round}
  rax:=rax*12/pi; {make hours}
  h:=trunc(rax);
  m:=trunc((rax-h)*60);
  s:=trunc((rax-h-m/60)*3600);
  ds:=trunc((rax-h-m/60-s/3600)*36000);
  Str(trunc(h):2,b);
  prepare_ra:=b+sep+' '+leadingzero(m)+'  '+leadingzero(s)+'.'+char(ds+48);
end;
function prepare_dec(decx:double; sep:char):ansistring; {radialen to text, format 90d 00 00}
 var
   B : string[9];
   g,m,s  :integer;
   sign   : char;
begin {make from rax [0..pi*2] a text in array bericht. Length is 10 long}
  if decx<0 then sign:='-' else sign:='+';
  decx:=abs(decx)+pi/(360*60*60); {add half second to get correct rounding and not 7:60 results as with round}
  decx:=decx*180/pi; {make degrees}
  g:=trunc(decx);
  m:=trunc((decx-g)*60);
  s:=trunc((decx-g-m/60)*3600);
  Str(trunc(g):2,b);
  prepare_dec:=sign+b+sep+' '+leadingzero(m)+'  '+leadingzero(s);
end;

procedure rotate(rot,x,y :double;var  x2,y2:double);{rotate a vector point}
begin
  x2:=x * + sin(rot)+y*cos(rot );{ROTATION MOON AROUND CENTER OF PLANET}
  y2:=x * - cos(rot)+y*sin(rot );{SEE PRISMA WIS VADEMECUM BLZ 68}
end;

FUNCTION fnmodulo (x,range: double):double;
BEGIN
   {range should be 2*pi or 24 hours or 0 .. 360}
   x:=range *frac(X /range); {quick method for big numbers}
   if x<0 then x:=x+range;   {do not like negative numbers}
   fnmodulo:=x;
END;


function intensity2(x:tcolor):integer;
begin
  intensity2:=round((GetBValue(x)+getGvalue(x)+getRvalue(x))/3);{get red, green blue value as intensity}
end;
function intensityRGB(x:tcolor): byteX3;
var
   dummy: byteX3;
begin
  dummy[0]:=getRvalue(x);{get red, green blue value as intensity}
  dummy[1]:=getGvalue(x);
  dummy[2]:=getBvalue(x);
  intensityRGB:=dummy;
end;

procedure old_to_new_WCS;{ convert old WCS to new}
var
   sign        : integer;
begin
  cd1_1:=cdelt1*cos(crota2*pi/180); {note 2013 should be crota1 if skewed}
  if cdelt1>=0 then sign:=+1 else sign:=-1;
  cd1_2:=abs(cdelt2)*sign*sin(crota2*pi/180);{note 2013 should be crota1 if skewed}
  if cdelt2>=0 then sign:=+1 else sign:=-1;
  cd2_1:=-abs(cdelt1)*sign*sin(crota2*pi/180);
  cd2_2:= cdelt2*cos(crota2*pi/180);
end;

procedure new_to_old_WCS;
var
   sign        : integer;
begin
{ convert to old WCS. Based on draft 1988 , do not conversion article Alain Klotz, give sometimes zero CROTA}
  if (cd1_1*cd2_2-cd1_2*cd2_1)>=0 then sign:=+1 else sign:=-1;

  cdelt1:=sqrt(sqr(cd1_1)+sqr(cd2_1))*sign;{if no old wcs header use cd2_2 of new WCS style for pixel size}
  cdelt2:=sqrt(sqr(cd1_2)+sqr(cd2_2));{if no old wcs header use cd2_2 of new WCS style for pixel size}

  crota2:= atn_2(sign*cd1_2,cd2_2);
  {note 2013 CROTA1=arctan2(-CD1_2,CD2_2)}
  {note 2013 CROTA2=arctan2(CD1_1,CD2_1)}
  crota2:= crota2*180/pi;{convert now radians to degrees as standard}
end;


Function INT_IEEE4(x:longword):single;{adapt non-intel floating to intel floating point}
var lw       : longword;
    fl       : single absolute lw;
begin
  lw:=(x and $FF) shl 24 + (x and $FF00) shl 8 + (x and $FF0000) shr 8 +(x and $FF000000) shr 24 ;
      {swap 4 bytes such that 1234 becomes 4321}
  INT_IEEE4:=fl;{get final floating value}
end;
Function SWAP32fast(x:longword):longword;{it very little faster then SWAP32 and SWAP32first,  adapt non-intel 32  bit  longword to intel longword   or big endian to little endian}
asm
  bswap eax
end;

Function INT_IEEE8(x: int64):double;{adapt non-intel floating double to intel floating double}
var lw       : int64;
    fl       : double absolute lw;
begin
  lw:=(x and $FF) shl 56 + (x and $FF00) shl 40 + (x and $FF0000) shl 24 + (x and $FF000000) shl 8 +(x and $FF00000000) shr 8 +(x and $FF0000000000) shr 24 + (x and $FF000000000000) shr 40 +(x and $FF00000000000000) shr 56;
      {swap 8 bytes such that 12345678 becomes 87654321}
  INT_IEEE8:=fl;{get final floating value}
end;


procedure loadfits(filen:ansistring);{load fits file, by Han Kleijn}
var
   i,j,k       : integer;
   col         : integer;
   s1          : ansistring;
   datamin     : integer;
   datamax     : integer;
   x,fract     : double;
   bzero       : integer;{zero shift. For example used in AMT}
   wo     : word;  {for 16 signed integer}
   sign_int  : smallint absolute wo;{for 16 signed integer}
   oldcd1_1,oldcd1_2,oldcd2_1,oldcd2_2 :double;
   rgbdummy    : byteX3;
   dimensions,dummy,naxis1  : integer;
  Function validate_double:double;{read values}
  var t :shortstring;
     r,err : integer;
     x   :double;
  begin
    t:='';
    for r:=10 to 29 do {read 20 characters max, position 11 to 30 in string, position 10 to 29 in pchar}
    if header[r] in [' ','/']=false then t:=t+header[r];{'/' is strictly not neccesary but safer}
    val(t,x,err);
    validate_double:=x;
   end;
   Function get_string:ansistring;{read values}
   var t :shortstring;
     r : integer;
   begin
     t:='';
     r:=11;
     repeat
       t:=t+header[r];
       inc(r);
     until ((header[r]=#39) or (r>=79+1));
     while t[length(t)] in [' ',#39] do setlength(t,length(t)-1);{remove spaces}
     get_string:=t;
   end;
   Function get_comment:string;{read values}
   var t :shortstring;
     r : integer;
   begin
     t:='';
     r:=10;
     repeat
       t:=t+header[r];
       inc(r);
     until (r>=79+1);
     while t[length(t)]=' ' do setlength(t,length(t)-1);{remove spaces}
     get_comment:=t;
   end;
   Function pos2(x:integer;s1:ansistring):ansistring;{get substring from string separated by kommas from position x}
   var k,l :integer;
     s2 : ansistring;
   begin
     l:=1;
     s2:='';
     for k:=1 to length(s1) do
     begin
       if s1[k]=',' then inc(l)
       else
       if l=x then s2:=s2+s1[k];{extract substring which is between kommas number x and x+1}
     end;
     pos2:=s2;
   end;
begin
  try
    TheFile3:=tfilestream.Create( filen, fmOpenRead );
  except
    beep;
    textout(realsky.image1.canvas.handle,30,30,'Error loading FITS file !',25);
    exit;
  end;
  realsky.Caption:=(filen);

  Reader := TReader.Create (theFile3,$4000);{number of hnsky records}
  {thefile3.size-reader.position>sizeof(hnskyhdr) could also be used but slow down a factor of 2 !!!}
  //I:=0;
  crota2:=0;{just for the case it is not available}
  ra0:=999;
  dec0:=0;
  cdelt1:=0;
  cdelt2:=0;
  bzero:=0;{just for the case it is not available}
  cd1_1:=0;{just for the case it is not available}
  cd1_2:=0;{just for the case it is not available}
  cd2_1:=0;{just for the case it is not available}
  cd2_2:=0;{just for the case it is not available}
  dimensions:=1;
  naxis1:=100;
  datamin:=0;datamax:=255; {if not defined, check later}

  repeat
    try reader.read(header[0],5*16); {read file info, 5*16 bytes only}
      except
        beep;
        Reader.free; TheFile3.free; exit;
      end;
    if ((header[0]='B') and (header[1]='I') and (header[2]='T') and (header[3]='P') and (header[4]='I') and (header[5]='X')) then
    begin
      nrbits:=round(validate_double);{BITPIX, read integer using double routine}
    end;
    if ((header[0]='N') and (header[1]='A') and (header[2]='X') and (header[3]='I') and (header[4]='S')) then
    begin
      if (header[5]=' ') then naxis:=round(validate_double) else {number of dimensions}
      if (header[5]='1') then begin
                                naxis1:=round(validate_double);
                                wide2:=naxis1;
                           end else    {NAXIS1 pixels}
      if (header[5]='2') then height2:=round(validate_double) else  {NAXIS2 pixels}
      if (header[5]='3') then
      begin {RGB fits, here to prevent error with setting references in window after change decD....}
        dummy:=round(validate_double);  {NAXIS2 pixels}
        if naxis1=3 then  {type NAXIS = 3 / Number of dimensions
                               NAXIS1 = 3 / Number of Colors
                               NAXIS2 = 382 / Row length
                               NAXIS3 = 255 / Number of rows}
        begin             {RGB fits with naxis1=3, treated as 24 bits coded pixels in 2 dimensions}
          wide2:=height2;
          height2:=round(validate_double);  {NAXIS2 pixels}
        end
        else {The alternative where the planes are in RGB-order using NAXIS1 = 382, NAXIS2 = 255 and  NAXIS3=3}
        begin
          dimensions:=dummy;{number of colors}
        end;
      end;
    end;
    if ((header[0]='D') and (header[1]='A') and (header[2]='T') and (header[3]='A') and (header[4]='M')) then
    begin
      if ((header[5]='I') and (header[6]='N')) then datamin:=round(validate_double) else {DATAMIN. Note some FIT files contain double}
      if ((header[5]='A') and (header[6]='X')) then datamax:=round(validate_double);{DATAmax. Note some FIT files contain double }
    end;
    if ((header[0]='B') and (header[1]='Z') and (header[2]='E') and (header[3]='R') and (header[4]='O') ) then
       bzero:=round(validate_double);
    if ((header[0]='C') and (header[1]='R') and (header[2]='O') and (header[3]='T') and (header[4]='A') {and (header[i+5]='2')}) then
       crota2:=validate_double;{DATAMIN, read double value}                          {last for r Ledas}

    if ((header[0]='C') and (header[1]='R') and (header[2]='P') and (header[3]='I') and (header[4]='X')) then
    begin
      if header[5]='1' then crpix1:=validate_double else{ref pixel for x}
      if header[5]='2' then crpix2:=validate_double;  {ref pixel for y}
    end;
    if ((header[0]='C') and (header[1]='D') and (header[2]='E') and (header[3]='L') and (header[4]='T')) then
    begin
      if header[5]='1' then cdelt1:=validate_double else{deg/pixel for RA}
      if header[5]='2' then cdelt2:=validate_double;  {deg/pixel for DEC}
    end;

    if ((header[0]='C') and (header[1]='R') and (header[2]='V') and (header[3]='A') and (header[4]='L')) then
    begin
      if (header[5]='1') then ra0:=validate_double*pi/180; {ra center, read double value}
      if (header[5]='2') then dec0:=validate_double*pi/180; {dec center, read double value}
    end;
    if ((header[0]='C') and (header[1]='D')) then
    begin
      if ((header[2]='1') and (header[3]='_') and (header[4]='1')) then   cd1_1:=validate_double;
      if ((header[2]='1') and (header[3]='_') and (header[4]='2')) then   cd1_2:=validate_double;
      if ((header[2]='2') and (header[3]='_') and (header[4]='1')) then   cd2_1:=validate_double;
      if ((header[2]='2') and (header[3]='_') and (header[4]='2')) then   cd2_2:=validate_double;
    end;

  until ((header[0]='E') and (header[1]='N') and (header[2]='D'));


  if ((naxis=3) and (naxis1=3)) then nrbits:=24; {threat RGB fits as 2 dimensional with 24 bits data}
  if cdelt2=0 then
    new_to_old_WCS {new WCS, convert to old WCS, required to show Crota and so on... Ignore skew info in CD matrix}
  else
  if cd1_1=0  then
  old_to_new_WCS;{ {new WCS missing, convert old WCS to new}


  angle:=(-crota2)*pi/180;

  fract:=frac(reader.position/2880);
  if fract<>0 then
  begin
    i:=round((1-fract)*2880);{left part of next 2880 bytes block. For realsky read total 4 x 2880 blocks}
    try reader.read(header[0],i);except;end; {skip empty part and go to image data}
  end;

  i:=round(bufwide/(abs(nrbits/8)));
  if wide2>i then
  begin
    beep;
    realsky.image1.canvas.textout(30,30,'Too large FITS file !');
    mainwindow.SaveasBMP1.enabled:=false;
  end
  else
  if ((nrbits=8) or (nrbits=16) or (nrbits=-32) or (nrbits=24) )= false  then
  begin
    beep;
    realsky.image1.canvas.textout(30,30,'BITPIX= '+inttostr(nrbits)+' not supported !');
  end
  else
  if ((naxis1=3) and (nrbits<>24)) then
  begin
    beep;
    realsky.image1.canvas.textout(30,30,'FITS file not supported !');
  end
  else
  begin {not too large and readable}
    realsky.clientwidth:=wide2; {window size}
    realsky.ClientHeight:=height2+realsky.statusbar1.ClientHeight; {window size}

    for k:=1 to dimensions do {for monochrome dimension is 1}
    begin
      for i:=0 to height2-1 do
      begin
        try reader.read(fitsbuffer,wide2*round(abs(nrbits/8)));except end; {read file info}

        if i=0 then
        begin {first line}
          if ((nrbits<>8) and (datamax=255)) then {datamin and max not in header, estimate using first image line}
          for j:=0 to wide2-1 do
          begin {find range}
            bzero:=0;
            case nrbits of  16: begin wo:=swap(fitsbuffer2[j]);;{move data to wo and therefore sign_int} col:=sign_int; end;
//                             32: col:=round(SWAP32fast(fitsbuffer4[j]));
                            -32: col:=round(INT_IEEE4(fitsbuffer4[j]));
                            -64: col:=round(INT_IEEE8(fitsbuffer8[j]));
                          end;{case}
            if col<datamin then datamin:=col else if col>datamax then datamax:=col;
          end; {find range}
        end; {first line}

        for j:=0 to wide2-1 do
        begin
          if nrbits=8 then col:=round( ((fitsbuffer[j]-datamin+bzero)*256/(datamax-datamin) )){scale to maximum contrast}
          else
          if nrbits=24 then rgbdummy:=fitsbufferRGB[j]{RGB fits with naxis1=3, treated as 24 bits coded pixels in 2 dimensions}
          else
          if nrbits=16 then
          begin
            wo:=swap(fitsbuffer2[j]);{move data to wo and therefore sign_int}
            col:=round( ((sign_int-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
          end
          else {4 byte floating point FITS image}
          if nrbits=-32 then col:=round( (( INT_IEEE4(fitsbuffer4[j])-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
             {int_IEEE, swap four bytes and the read as floating point}

          if col>255 then col:=255;
          if col<0 then col:=0;

          if naxis<>3 then
          begin   {monochrome FITS}
            try setpixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(col,col,col));except;end;
          end
          else

          begin {RGB fits}
            if nrbits=24 then {RGB fits with naxis1=3, treated as 24 bits coded pixels in 2 dimensions or 24 bits FITS}
            begin
              try setpixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(rgbdummy[0],rgbdummy[1],rgbdummy[2]) );except;end;
            end {RGB fits with naxis1=3}
            else

            begin {3 dimensions RGB, naxis3=3, first dimensions is red, second is green and so on}
              case k of 1:  try setpixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(col,0,0) );except;end; {RED} {first time do not read pixel, only seond and third run}
                        2:  begin
                                 try dummy:=getpixel(realsky.image1.canvas.handle,j,(height2-1)-i);
                                 setPixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(getrvalue(dummy),col,0) ) ; {add GREEN}
                                 except;end;
                            end;
                        3:  begin
                                try dummy:=getpixel(realsky.image1.canvas.handle,j,(height2-1)-i);
                                setPixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(getRvalue(dummy),getGvalue(dummy),col) ) ;  {add BLUE}
                                except;end;
                            end;
               end;{case}
            end;{RGB naxis3=3}
          end;{RGB}
        end;{j}
      end;{i}
    end;{k}
  end;{not too large and supported}

  Reader.free;
  TheFile3.free;

  invalidaterect(realsky.handle,nil,false);{paint again}
end;

procedure plotfits(filen:ansistring);{plot FITS scaled in 360x180 degrees Equirectangular projection}
var
   sum_of_values, nr_of_values  : array of array of integer;{{dynamic 2 dimensions array}
   i,j,k       : integer;
   col         : integer;
   s1          : ansistring;
   datamin     : integer;
   datamax     : integer;
   x,fract     : double;
   bzero       : integer;{zero shift. For example used in AMT}
   wo     : word;  {for 16 signed integer}
   dRa,dDec,gamma ,sin_dec0,cos_dec0,delta  : double;
   sign_int  : smallint absolute wo;{for 16 signed integer}
   oldcd1_1,oldcd1_2,oldcd2_1,oldcd2_2 :double;
   rgbdummy    : byteX3;
   dimensions,dummy,naxis1, fits_width, fits_height,x1,y1: integer;
  Function validate_double:double;{read values}
  var t :shortstring;
     r,err : integer;
     x   :double;
  begin
    t:='';
    for r:=I+10 to I+29 do {read 20 characters max, position 11 to 30 in string, position 10 to 29 in pchar}
    if header[r] in [' ','/']=false then t:=t+header[r];{'/' is strictly not neccesary but safer}
    val(t,x,err);
    validate_double:=x;
   end;
   Function get_string:ansistring;{read values}
   var t :shortstring;
     r : integer;
   begin
     t:='';
     r:=I+11;
     repeat
       t:=t+header[r];
       inc(r);
     until ((header[r]=#39) or (r>=I+79+1));
     while t[length(t)] in [' ',#39] do setlength(t,length(t)-1);{remove spaces}
     get_string:=t;
   end;
   Function get_comment:ansistring;{read values}
   var t :shortstring;
     r : integer;
   begin
     t:='';
     r:=I+10;
     repeat
       t:=t+header[r];
       inc(r);
     until (r>=I+79+1);
     while t[length(t)]=' ' do setlength(t,length(t)-1);{remove spaces}
     get_comment:=t;
   end;
   Function pos2(x:integer;s1:ansistring):ansistring;{get substring from string separated by kommas from position x}
   var k,l :integer;
     s2 : ansistring;
   begin
     l:=1;
     s2:='';
     for k:=1 to length(s1) do
     begin
       if s1[k]=',' then inc(l)
       else
       if l=x then s2:=s2+s1[k];{extract substring which is between kommas number x and x+1}
     end;
     pos2:=s2;
   end;
begin
  try
    TheFile3:=tfilestream.Create( filen, fmOpenRead );
  except
    beep;
    textout(realsky.image1.canvas.handle,30,30,'Error loading FITS file !',25);
    exit;
  end;
//  realsky.caption:=ExtractFileName(filen);

  Reader := TReader.Create (theFile3,$4000);{number of hnsky records}
  {thefile3.size-reader.position>sizeof(hnskyhdr) could also be used but slow down a factor of 2 !!!}
  I:=0;
  crota2:=0;{just for the case it is not available}
  ra0:=999;
  dec0:=0;
  cdelt1:=0;
  cdelt2:=0;
  bzero:=0;{just for the case it is not available}
  cd1_1:=0;{just for the case it is not available}
  cd1_2:=0;{just for the case it is not available}
  cd2_1:=0;{just for the case it is not available}
  cd2_2:=0;{just for the case it is not available}
  dimensions:=1;
  naxis1:=100;
  bzero:=0;
  datamin:=0;datamax:=255; {if not defined, check later}

  repeat
    try reader.read(header[0],5*16); {read file info, 5*16 bytes only}
      except
        beep;
        Reader.free; TheFile3.free; exit;
      end;
    if ((header[0]='B') and (header[1]='I') and (header[2]='T') and (header[3]='P') and (header[4]='I') and (header[5]='X')) then
    begin
      nrbits:=round(validate_double);{BITPIX, read integer using double routine}
    end;
    if ((header[0]='N') and (header[1]='A') and (header[2]='X') and (header[3]='I') and (header[4]='S')) then
    begin
      if (header[5]=' ') then naxis:=round(validate_double) else {number of dimensions}
      if (header[5]='1') then begin
                                naxis1:=round(validate_double);
                                fits_width:=naxis1;
                           end else    {NAXIS1 pixels}
      if (header[5]='2') then fits_height:=round(validate_double) else  {NAXIS2 pixels}
      if (header[5]='3') then
      begin {RGB fits, here to prevent error with setting references in window after change decD....}
        dummy:=round(validate_double);  {NAXIS2 pixels}
        if naxis1=3 then  {type NAXIS = 3 / Number of dimensions
                               NAXIS1 = 3 / Number of Colors
                               NAXIS2 = 382 / Row length
                               NAXIS3 = 255 / Number of rows}
        begin             {RGB fits with naxis1=3, treated as 24 bits coded pixels in 2 dimensions}
          fits_width:=fits_height;
          fits_height:=round(validate_double);  {NAXIS2 pixels}
        end
        else {The alternative where the planes are in RGB-order using NAXIS1 = 382, NAXIS2 = 255 and  NAXIS3=3}
        begin
          dimensions:=dummy;{number of colors}
        end;
      end;
    end;
    if ((header[0]='D') and (header[1]='A') and (header[2]='T') and (header[3]='A') and (header[4]='M')) then
    begin
      if ((header[5]='I') and (header[6]='N')) then datamin:=round(validate_double) else {DATAMIN. Note some FIT files contain double}
      if ((header[5]='A') and (header[6]='X')) then datamax:=round(validate_double);{DATAmax. Note some FIT files contain double }
    end;
    if ((header[0]='B') and (header[1]='Z') and (header[2]='E') and (header[3]='R') and (header[4]='O') ) then
       bzero:=round(validate_double);
    if ((header[0]='C') and (header[1]='R') and (header[2]='O') and (header[3]='T') and (header[4]='A') {and (header[i+5]='2')}) then
       crota2:=validate_double;{DATAMIN, read double value}                          {last for r Ledas}

    if ((header[0]='C') and (header[1]='R') and (header[2]='P') and (header[3]='I') and (header[4]='X')) then
    begin
      if header[5]='1' then crpix1:=validate_double else{ref pixel for x}
      if header[5]='2' then crpix2:=validate_double;  {ref pixel for y}
    end;
    if ((header[0]='C') and (header[1]='D') and (header[2]='E') and (header[3]='L') and (header[4]='T')) then
    begin
      if header[5]='1' then cdelt1:=validate_double else{deg/pixel for RA}
      if header[5]='2' then cdelt2:=validate_double;  {deg/pixel for DEC}
    end;

    if ((header[0]='C') and (header[1]='R') and (header[2]='V') and (header[3]='A') and (header[4]='L')) then
    begin
      if (header[5]='1') then ra0:=validate_double*pi/180; {ra center, read double value}
      if (header[5]='2') then dec0:=validate_double*pi/180; {dec center, read double value}
    end;
    if ((header[0]='C') and (header[1]='D')) then
    begin
      if ((header[2]='1') and (header[3]='_') and (header[4]='1')) then   cd1_1:=validate_double;
      if ((header[2]='1') and (header[3]='_') and (header[4]='2')) then   cd1_2:=validate_double;
      if ((header[2]='2') and (header[3]='_') and (header[4]='1')) then   cd2_1:=validate_double;
      if ((header[2]='2') and (header[3]='_') and (header[4]='2')) then   cd2_2:=validate_double;
    end;

  until ((header[0]='E') and (header[1]='N') and (header[2]='D'));

  if cd1_1=0  then
  old_to_new_WCS;{ {new WCS missing, convert old WCS to new}

  fract:=frac(reader.position/2880);
  if fract<>0 then
  begin
    i:=round((1-fract)*2880);{left part of next 2880 bytes block. For realsky read total 4 x 2880 blocks}
    try reader.read(header[0],i);except;end; {skip empty part and go to image data}
  end;

//  i:=round(bufwide/(abs(nrbits/8)));
//  if wide2>i then
//  begin
//    beep;
//    realsky.image1.canvas.textout(30,30,'Too large FITS file !');
//    mainwindow.SaveasBMP1.enabled:=false;
//  end
//  else
  if ((nrbits=8) or (nrbits=16) or (nrbits=-32) )= false  then
  begin
    beep;
  end
  else
  begin {not too large and readable}
//   if nrbits=-32 then {skyview}
//   begin
//     datamin:=4500;datamax:=14000; {if not defined, check later}
//   end;

    SetLength(sum_of_values, (height2), wide2);{set size dynamic array}
    SetLength(nr_of_values, height2, wide2);
    for i:=0 to height2-1 do {range rect projection}
    begin
      for j:=0 to wide2-1 do
      begin {empthy values}

        sum_of_values[i,j]:=0;
        nr_of_values[i,j]:=0;
      end;
    end;

      for i:=0 to (fits_height-1) do  {range FITS size}
      begin
        try reader.read(fitsbuffer,fits_width*round(abs(nrbits/8)));except end; {read file info}

       if i=0 then
        begin {first line}
          if ((nrbits<>8) and (datamax=255)) then {datamin and max not in header, estimate using first image line}
          for j:=0 to fits_height-1 do
          begin {find range}
            bzero:=0;
           case nrbits of  16: begin wo:=swap(fitsbuffer2[j]);;{move data to wo and therefore sign_int} col:=sign_int; end;
                             32: col:=round(SWAP32fast(fitsbuffer4[j]));
                            -32: col:=round(INT_IEEE4(fitsbuffer4[j]));
                            -64: col:=round(INT_IEEE8(fitsbuffer8[j]));
                          end;{case}
            if col<datamin then datamin:=col else if col>datamax then datamax:=col;
          end; {find range}
        end; {first line}


        for j:=0 to (fits_width-1) do
        begin
          if nrbits=8 then col:=round( ((fitsbuffer[j]-datamin+bzero)*256/(datamax-datamin) )){scale to maximum contrast}
          else
          if nrbits=16 then
          begin
            wo:=swap(fitsbuffer2[j]);{move data to wo and therefore sign_int}
            col:=round( ((sign_int-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
            //col:=round(sqrt(col+1)*sqrt(255));
          end
          else {4 byte floating point FITS image}
          if nrbits=-32 then col:=round( (( INT_IEEE4(fitsbuffer4[j])-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
             {int_IEEE, swap four bytes and the read as floating point}


          //col:=col-100;

          if col>255 then col:=255;
          if col<0 then col:=0;

          dRa :=(cd1_1*(j+1-crpix1)+cd1_2*(i+1-crpix2))*pi/180;{+1 because pixels are counted in FITS files from 1 to wide/length !!}
          dDec:=(cd2_1*(j+1-crpix1)+cd2_2*(i+1-crpix2))*pi/180;
          sincos(dec0,sin_dec0,cos_dec0);
          delta:=cos_dec0-dDec*sin_dec0;
          gamma:=sqrt(dRa*dRa+delta*delta);
          RAM:=RA0+arctan(Dra/delta);
          DECM:=arctan((sin_dec0+dDec*cos_dec0)/gamma);
            {new WCS}
          if RAM<0 then RAM:=RAM+pi*2;

          x1:=round(( wide2-1)*(1-ram/(2*pi)));
          y1:=round((height2-1)*(0.5-decm/pi));

            {sum up values}
          sum_of_values[y1,x1]:=sum_of_values[y1,x1]+col;
          nr_of_values[y1,x1]:=nr_of_values[y1,x1]+1
        end;{j}
      end;{i}
  end;{not too large and supported}

  for i:=0 to height2-1 do {range rect}
  begin
    for j:=0 to wide2-1 do
    begin {average value to pixel}
      if nr_of_values[i,j]<>0 then
      begin
        col:=round(sum_of_values[i,j]/nr_of_values[i,j]);
        try setpixel(realsky.image1.canvas.handle,j,i,rgb(col,col,col));except;end;
      end;
    end;
  end;
  nr_of_values := nil;{remove dynamic array from memeory}
  sum_of_values:=nil;{remove dynamic array from memeory}
  Reader.free;
  TheFile3.free;

  invalidaterect(realsky.handle,nil,false);{paint again}
end;





procedure compensate_strech(x,y: double;var x2,y2:double);
begin
 {correct mouse position for possible stretched image, input and output range 0....}
 {example 1 pixel nr 2 in range 1 to 4 is in the middle of pixel 2 or 2-0.5 equals 1.5/4 or 37.5% }
 {   if range is doubled, pixel 2 becomes pixel 3 and 4. Position is between pixel 3 and 4
    equals 3/8 or 37.5%
    if range is trippled, pixel 2 becomes pixel 4,5,6. Position halverwege pixel 5 is 5-0.5 or 4.5 equals 4.5/12 =37.5%}
 x2:=(x+1-0.5)*(wide2)/(realsky.image1.width);
 y2:=(y+1-0.5)*(height2)/(realsky.image1.height);
 x2:=x2-1+0.5; {change range back to 0...}
 y2:=y2-1+0.5; {change range back to 0...}
end;

procedure de_compensate_strech(x,y: double;var x2,y2:double); {correct mouse position for possible stretched image}
begin
 {correct fits position for possible stretched image, input and output range 0....}

 x2:=round((x+1-0.5)*(realsky.image1.width)/(wide2));
 y2:=round((y+1-0.5)*(realsky.image1.height)/(height2));
 x2:=x2-1+0.5; {change range back to 0...}
 y2:=y2-1+0.5; {change range back to 0...}

end;


procedure Trealsky.Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
 Y: Integer);
var
 xx1,yy1,yy,xx,arcsinx,mx1,my1,dRa,dDec,delta,gamma : double;
 s1,s2,s3, title2: string;
 fitsx,fitsy : double;
 decimalen :integer;
begin
  compensate_strech(x,y,mx1,my1);{correct mouse position for possible stretched image}

  if nrbits<>-8 then
  begin {FITS file}
    fitsx:=mx1+1;{starts at 1}
    fitsy:=height2-my1; {from bottom to top, starts at 1}

    dRa :=(cd1_1*(fitsx-crpix1)+cd1_2*(fitsy-crpix2))*pi/180;
    dDec:=(cd2_1*(fitsx-crpix1)+cd2_2*(fitsy-crpix2))*pi/180;
    delta:=cos(dec0)-dDec*sin(dec0);
    gamma:=sqrt(dRa*dRa+delta*delta);
    RAM:=RA0+arctan(Dra/delta);
    DECM:=arctan((sin(dec0)+dDec*cos(dec0))/gamma);

    RAM:=fnmodulo (RAM,2*pi);

    if ((height2<realsky.image1.height) or (wide2<realsky.image1.width)) then decimalen:=1 else decimalen:=0;
    str(fitsx:0:decimalen,s1); {fits images start with 1 and not with 0}
    str(fitsy:0:decimalen,s2); {Y from bottom to top}
    title2:='FITS: ';
  end
  else
  if wide2<>height2 then
  begin {{rectangular projection, not a FITS files}

    RAM:=2*pi*(1- mx1/(wide2));
    DECM:=pi*(0.5 - my1/(height2));

    if ((height2<realsky.image1.height) or (wide2<realsky.image1.width)) then decimalen:=1 else decimalen:=0;
    str(mx1:0:decimalen,s1);
    str(my1:0:decimalen,s2);
    title2:='Equirectangular projection:  ';

  end
  else
  begin {toast projection}
    de_toast(round(mx1), round(my1), height2, ram,decm); {de-TOAST x,y coordinates to lat, lon. The origin (0,0) is top left of TOAST projection, size:=height or witdh}
    if ((height2<realsky.image1.height) or (wide2<realsky.image1.width)) then decimalen:=1 else decimalen:=0;
    str(mx1:0:decimalen,s1);
    str(my1:0:decimalen,s2);
    title2:='TOAST projection: ';
  end;

  try str(intensity2(getpixel(realsky.image1.canvas.handle,round(mx1),round(my1))):4,s3);except;end;

  statusbar1.simpletext:=title2+s3+'=('+s1+', '+s2+')     '+prepare_ra(raM,':')+'     '+prepare_dec(decM,'d');
  realsky.Caption:=filename_short+'    '+title2+s3+'=('+s1+', '+s2+')     '+prepare_ra(raM,':')+'     '+prepare_dec(decM,'d')
end;

procedure Trealsky.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 mainwindow.SaveasBMP1.enabled:=false;
end;

procedure Trealsky.FormShow(Sender: TObject);
begin
 mainwindow.SaveasBMP1.enabled:=true;
end;

Function strtofloat2(s : ansistring):double;
var
 y : double;
 error: integer;
begin
 {For Window version where floatings are written with a komma, the standard
strtofloat does not accept 1.33 but only 1,33}
 {So this is an alternative routine with the good old Val procedure}
 val(s,y,error);
 if error<>0 then raise Exception.Create(s+' contains illegal character(s)!');
 strtofloat2:=y;
end;

Function deg_and_minutes_tofloat(s:ansistring):double;
var
 x: double;
 j: integer;
begin
 j:=pos(':',s);
 if j=0 then {12.50 format}
  x:=strtofloat2(s) {12.5 format}
 else {12:30.0 format}
 begin
  x:=(strtofloat2(copy(s,1,j-1)));
  if pos('-',s)>0 then x:=x - strtofloat2(copy(s,j+1,length(s)-j))/60
      else x:=x + strtofloat2(copy(s,j+1,length(s)-j))/60 ;

 end;
 deg_and_minutes_tofloat:=x;
end;



PROCEDURE ang_sep(ra1,dec1,ra2,dec2 : double;var sep: double); { By Han Kleijn}
var cos_sep:double;
{caculates angular separation. according formula 9.1 old meeus}
Begin
 cos_sep:=sin(dec1)*sin(dec2)+ cos(dec1)*cos(dec2)*cos(ra1-ra2);
 if ABS(cos_sep)<1 then begin sep:=(PI/2-ARCTAN(cos_sep/(SQRT(1-SQR(cos_sep))))) end {arccos function}
 else
 begin if cos_sep>0 then sep:=0 {note bij x=1 klapt uitkomst net om}
      else sep:=PI;
 end; {arccos function}

end;



{$R hns_fits_cursor.res} {FOR CURSORS}
const
 crMyCursor = 5;
procedure Trealsky.FormCreate(Sender: TObject);
begin
 Screen.Cursors[crMyCursor] := LoadCursor(HInstance, 'CROSS_CURSOR');
 Cursor := crMyCursor;
end;

procedure Trealsky.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
//  realsky.height:=round(realsky.image1.height*0.9);
//  realsky.width:=round(realsky.image1.width*0.9);
  realsky.height:=round(realsky.height*0.9);
  realsky.width:=round(realsky.width*0.9);

end;

procedure Trealsky.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
//  realsky.image1.height:=round(realsky.image1.height*1.1);
//  realsky.image1.width:=round(realsky.image1.width*1.1);
  realsky.height:=round(realsky.height*1.1);
  realsky.width:=round(realsky.width*1.1);
end;

procedure Trealsky.Image1MouseDown(Sender: TObject; Button: TMouseButton;
 Shift: TShiftState; X, Y: Integer);
var
 s   : ansistring;
 i,j,col  : integer;
 fitsx,fitsy ,max_x,max_y,xs,ys :double;
 cc2 :tpoint;
 tot_col, k : integer;
begin
 xs:=x;
 ys:=y;
 compensate_strech(xs,ys,xs,ys); {correct mouse position for possible stretched image}

 fitsx:=(xs)+1;{starts at 1}
 fitsy:=height2-(ys); {from bottom to top, starts at 1} {caculated for no-stretched image}


 cc2:=realsky.Image1.ClientOrigin;
 de_compensate_strech(xs,ys,xs,ys);
 setcursorpos(round(cc2.x+xs),round(cc2.y+ys));
end;

end.
